/* PulseSourceDataLineTest.java
   Copyright (C) 2008 Red Hat, Inc.

This file is part of IcedTea.

IcedTea is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2.

IcedTea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with IcedTea; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
 */

package org.classpath.icedtea.pulseaudio;

import java.io.File;
import java.io.IOException;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.BooleanControl;
import javax.sound.sampled.Control;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.FloatControl;
import javax.sound.sampled.Line;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.UnsupportedAudioFileException;

import junit.framework.JUnit4TestAdapter;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

public class PulseSourceDataLineTest {
	Mixer mixer;

	private int listenerCalled = 0;

	int started = 0;
	int stopped = 0;

	AudioFormat aSupportedFormat = new AudioFormat(
			AudioFormat.Encoding.PCM_UNSIGNED, 44100f, 8, 1, 1, 10, true);

	public static junit.framework.Test suite() {
		return new JUnit4TestAdapter(PulseSourceDataLineTest.class);
	}

	@Before
	public void setUp() throws Exception {
		mixer = PulseAudioMixer.getInstance();
		mixer.open();

		started = 0;
		stopped = 0;

	}

	@Test
	public void testOpenAndClose() throws LineUnavailableException {
		SourceDataLine line = (SourceDataLine) mixer.getLine(new Line.Info(
				SourceDataLine.class));

		line.open();
		Assert.assertTrue(line.isOpen());

		line.close();
		Assert.assertFalse(line.isOpen());

	}

	@Test
	public void testIsActiveAndIsOpen() throws LineUnavailableException {

		SourceDataLine line = (SourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, aSupportedFormat, 1000));

		Assert.assertFalse(line.isActive());
		Assert.assertFalse(line.isOpen());
		line.open();
		Assert.assertTrue(line.isOpen());
		Assert.assertFalse(line.isActive());
		line.start();
		Assert.assertTrue(line.isOpen());
		Assert.assertTrue(line.isActive());
		line.stop();
		Assert.assertTrue(line.isOpen());
		Assert.assertFalse(line.isActive());
		line.close();
		Assert.assertFalse(line.isOpen());
		Assert.assertFalse(line.isActive());

	}

	@Test
	public void testPlay() throws LineUnavailableException,
			UnsupportedAudioFileException, IOException {
		System.out.println("This test plays a file");

		File soundFile = new File("testsounds/startup.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		AudioFormat audioFormat = audioInputStream.getFormat();

		SourceDataLine line;
		line = (PulseAudioSourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, audioFormat));
		Assert.assertNotNull(line);

		line.open(audioFormat);
		System.out.println("opened");
		line.start();
		System.out.println("started");
		byte[] abData = new byte[1000];
		int bytesRead = 0;

		while (bytesRead >= 0) {
			bytesRead = audioInputStream.read(abData, 0, abData.length);
			if (bytesRead > 0) {
				line.write(abData, 0, bytesRead);
			}
		}
		System.out.println("done");

		line.drain();
		System.out.println("drained");
		line.stop();
		line.close();
		System.out.println("closed");

	}

	@Test
	public void testStartedStopped() throws LineUnavailableException,
			UnsupportedAudioFileException, IOException {

		File soundFile = new File("testsounds/startup.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		AudioFormat audioFormat = audioInputStream.getFormat();

		SourceDataLine line;
		line = (PulseAudioSourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, audioFormat));
		Assert.assertNotNull(line);

		line.open(audioFormat);

		LineListener startStopListener = new LineListener() {

			@Override
			public void update(LineEvent event) {
				if (event.getType() == LineEvent.Type.START) {
					started++;
					Assert.assertEquals(1, started);
				}

				if (event.getType() == LineEvent.Type.STOP) {
					System.out.println("Stopped event");
					stopped++;
					Assert.assertEquals(1, stopped);
				}
			}

		};

		line.addLineListener(startStopListener);

		byte[] abData = new byte[1000];
		int bytesRead = 0;

		line.start();
		while (bytesRead >= 0) {
			bytesRead = audioInputStream.read(abData, 0, abData.length);
			if (bytesRead > 0) {
				line.write(abData, 0, bytesRead);
			}
		}
		line.drain();

		line.stop();
		line.close();

		Assert.assertEquals(1, started);
		Assert.assertEquals(1, stopped);

	}

	@Test
	public void testStartNotificationOnCork()
			throws UnsupportedAudioFileException, IOException,
			LineUnavailableException {

		File soundFile = new File("testsounds/startup.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		AudioFormat audioFormat = audioInputStream.getFormat();

		PulseAudioSourceDataLine line;
		line = (PulseAudioSourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, audioFormat));
		Assert.assertNotNull(line);

		line.open(audioFormat);

		LineListener startStopListener = new LineListener() {

			@Override
			public void update(LineEvent event) {
				if (event.getType() == LineEvent.Type.START) {
					started++;
				}

				if (event.getType() == LineEvent.Type.STOP) {
					stopped++;
				}
			}

		};

		line.addLineListener(startStopListener);

		byte[] abData = new byte[1000];
		int bytesRead = 0;

		line.start();
		int count = 0;

		while (bytesRead >= 0) {
			bytesRead = audioInputStream.read(abData, 0, abData.length);
			if (bytesRead > 0) {
				line.write(abData, 0, bytesRead);
				count++;
				/*
				 * keep count high. if it is too low, the line wont even start
				 * playing so stopping is out of the question
				 */
				if (count == 100) {
					Operation o;
					synchronized (EventLoop.getEventLoop().threadLock) {
						o = line.getStream().cork();
					}

					o.waitForCompletion();
					o.releaseReference();

					try {
						Thread.sleep(1000);
					} catch (InterruptedException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}

					synchronized (EventLoop.getEventLoop().threadLock) {
						o = line.getStream().unCork();
					}

					o.waitForCompletion();
					o.releaseReference();

				}
			}
		}

		line.drain();

		line.stop();
		line.close();

		Assert.assertEquals(2, started);
		Assert.assertEquals(2, stopped);

	}

	@Test(expected = IllegalArgumentException.class)
	public void testPlayLessThanFrameSize() throws LineUnavailableException,
			UnsupportedAudioFileException, IOException {
		File soundFile = new File("testsounds/startup.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		AudioFormat audioFormat = audioInputStream.getFormat();
		// the audio file must have an even number of channels
		Assert.assertTrue(audioFormat.getChannels() % 2 == 0);
		SourceDataLine line;
		line = (SourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, audioFormat));

		byte[] data = new byte[1];
		data[0] = (byte) 'a';

		line.open();
		line.start();
		try {
			line.write(data, 0, 1);
		} finally {
			line.drain();
			line.stop();
			line.close();
		}
	}

	@Test
	public void testOpenFormat() throws LineUnavailableException,
			UnsupportedAudioFileException, IOException {
		/*
		 * This test makes sure that the default format of a line using open()
		 * is the same format that was passed to the mixer's getLine() function
		 * to get the line in the first place
		 */

		File soundFile = new File("testsounds/startup.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		AudioFormat audioFormat = audioInputStream.getFormat();

		SourceDataLine line;
		line = (SourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, audioFormat));
		Assert.assertNotNull(line);
		line.open();
		Assert.assertTrue(line.getFormat().matches(audioFormat));
		line.close();
	}

	@Test
	public void testFindLineWithFormat() throws LineUnavailableException {
		System.out
				.println("This test tries to find a line with a valid format");
		AudioFormat wantedFormat = new AudioFormat(
				AudioFormat.Encoding.PCM_UNSIGNED, 44100f, 8, 1, 1, 10, true);
		System.out.println(wantedFormat);
		SourceDataLine line = (SourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, wantedFormat));
		line.open();
		System.out.println(line.getFormat());
		line.close();

	}

	@Test(expected = IllegalArgumentException.class)
	public void testFindLineWithWrongFormat() throws LineUnavailableException {
		System.out
				.println("This test tries to acquire a line with incorrect format spec");
		SourceDataLine line = (SourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, new AudioFormat(
						AudioFormat.Encoding.PCM_UNSIGNED, 44100, 10000, 1, 13,
						10, true)));
		line.open();
		line.close();
	}

	@Test
	public void testVolumeAndMute() throws Exception {

		Mixer selectedMixer = mixer;
		SourceDataLine line = (SourceDataLine) selectedMixer
				.getLine(new Line.Info(SourceDataLine.class));

		File soundFile = new File(new java.io.File(".").getCanonicalPath()
				+ "/testsounds/logout.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		AudioFormat audioFormat = audioInputStream.getFormat();

		line.open(audioFormat);
		line.start();
		PulseAudioVolumeControl volume = (PulseAudioVolumeControl) line
				.getControl(FloatControl.Type.VOLUME);
		PulseAudioMuteControl mute = (PulseAudioMuteControl) line
				.getControl(BooleanControl.Type.MUTE);

		mute.setValue(true);
		volume.setValue(PulseAudioVolumeControl.MAX_VOLUME);

		mute.setValue(false);

		byte[] abData = new byte[1000];
		int bytesRead = 0;

		while (bytesRead >= 0) {
			bytesRead = audioInputStream.read(abData, 0, abData.length);
			if (bytesRead > 0) {
				line.write(abData, 0, bytesRead);
			}
		}

		line.drain();
		line.close();
		selectedMixer.close();

	}

	@Test
	public void testVolumeChanging() throws LineUnavailableException,
			IOException, UnsupportedAudioFileException {

		Mixer selectedMixer = mixer;

		SourceDataLine line = (SourceDataLine) selectedMixer
				.getLine(new Line.Info(SourceDataLine.class));

		File soundFile = new File(new java.io.File(".").getCanonicalPath()
				+ "/testsounds/logout.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		AudioFormat audioFormat = audioInputStream.getFormat();

		line.open(audioFormat);
		line.start();
		PulseAudioVolumeControl volume = (PulseAudioVolumeControl) line
				.getControl(FloatControl.Type.VOLUME);

		volume.setValue(PulseAudioVolumeControl.MIN_VOLUME);

		byte[] abData = new byte[1000];
		int bytesRead = 0;

		while (bytesRead >= 0) {
			bytesRead = audioInputStream.read(abData, 0, abData.length);
			if (bytesRead > 0) {
				line.write(abData, 0, bytesRead);
				volume.setValue(volume.getValue() + 100);
			}
		}

		line.drain();
		line.close();
		selectedMixer.close();

	}

	@Test
	public void testFindControl() throws LineUnavailableException {
		SourceDataLine sourceLine = (SourceDataLine) mixer
				.getLine(new Line.Info(SourceDataLine.class));
		sourceLine.open();
		Control[] controls = sourceLine.getControls();
		Assert.assertNotNull(controls);
		Assert.assertTrue(sourceLine.getControls().length > 0);
		for (Control control : controls) {
			Assert.assertNotNull(control);
		}
		sourceLine.close();
	}

	@Test
	public void testSupportedControls() throws LineUnavailableException {
		SourceDataLine sourceLine = (SourceDataLine) mixer
				.getLine(new Line.Info(SourceDataLine.class));
		sourceLine.open();
		Assert.assertTrue(sourceLine
				.isControlSupported(FloatControl.Type.VOLUME));
		Assert.assertTrue(sourceLine
				.isControlSupported(BooleanControl.Type.MUTE));
		sourceLine.close();
	}

	@Test
	public void testOpenEvent() throws LineUnavailableException {

		listenerCalled = 0;
		LineListener openListener = new LineListener() {
			public void update(LineEvent event) {
				Assert.assertTrue(event.getType() == LineEvent.Type.OPEN);
				PulseSourceDataLineTest.this.listenerCalled++;
			}
		};

		SourceDataLine line = (SourceDataLine) mixer.getLine(new Line.Info(
				SourceDataLine.class));
		line.addLineListener(openListener);
		line.open();
		line.removeLineListener(openListener);
		line.close();
		Assert.assertEquals(1, listenerCalled);
		listenerCalled = 0;
	}

	@Test
	public void testCloseEvent() throws LineUnavailableException {
		listenerCalled = 0;
		LineListener closeListener = new LineListener() {
			public void update(LineEvent event) {
				Assert.assertTrue(event.getType() == LineEvent.Type.CLOSE);
				PulseSourceDataLineTest.this.listenerCalled++;
			}
		};

		SourceDataLine line = (SourceDataLine) mixer.getLine(new Line.Info(
				SourceDataLine.class));
		line.open();
		line.addLineListener(closeListener);
		line.close();
		line.removeLineListener(closeListener);
		Assert.assertEquals(1, listenerCalled);
		listenerCalled = 0;
	}

	@Test
	public void testCloseEventWrongListener() throws LineUnavailableException {
		listenerCalled = 0;
		LineListener closeListener = new LineListener() {
			public void update(LineEvent event) {
				PulseSourceDataLineTest.this.listenerCalled++;
			}
		};

		SourceDataLine line = (SourceDataLine) mixer.getLine(new Line.Info(
				SourceDataLine.class));

		line.open();
		line.addLineListener(closeListener);
		line.removeLineListener(closeListener);
		line.close();
		Assert.assertEquals(0, listenerCalled);
		listenerCalled = 0;

	}

	@Test
	public void testFramePosition() throws UnsupportedAudioFileException,
			IOException, LineUnavailableException {
		File soundFile = new File("testsounds/logout.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		AudioFormat audioFormat = audioInputStream.getFormat();

		SourceDataLine line;
		line = (PulseAudioSourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, audioFormat));
		Assert.assertNotNull(line);

		line.open(audioFormat);
		line.start();

		byte[] abData = new byte[1000];
		int bytesRead = 0;

		while (bytesRead >= 0) {
			bytesRead = audioInputStream.read(abData, 0, abData.length);
			if (bytesRead > 0) {
				line.write(abData, 0, bytesRead);
			}
		}

		line.drain();
		line.stop();
		System.out.println("time position: " + line.getMicrosecondPosition());
		Assert.assertEquals(6199L, line.getMicrosecondPosition());
		line.close();
	}

	@Test
	public void testBufferSizes() throws LineUnavailableException {
		SourceDataLine line = (SourceDataLine) mixer.getLine(new Line.Info(
				SourceDataLine.class));
		line.open(aSupportedFormat, 10000);
		Assert.assertEquals(10000, line.getBufferSize());
		line.close();
	}

	@Test
	public void testHasADefaultFormat() throws LineUnavailableException {
		SourceDataLine line = (SourceDataLine) mixer.getLine(new Line.Info(
				SourceDataLine.class));
		Assert.assertNotNull(line.getFormat());
		System.out.println(line.getFormat());

	}

	@Test
	public void testDefaultFormatWithGetLine() throws LineUnavailableException {
		SourceDataLine line = (SourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, aSupportedFormat, 1000));
		Assert.assertEquals(aSupportedFormat, line.getFormat());

	}

	@Test
	public void testDefaultBufferSize() throws LineUnavailableException {
		SourceDataLine line = (SourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, aSupportedFormat, 1000));
		Assert.assertEquals(StreamBufferAttributes.SANE_DEFAULT, line
				.getBufferSize());
	}

	@Test
	public void testMixerKnowsAboutOpenLines() throws LineUnavailableException {
		SourceDataLine sourceDataLine = (SourceDataLine) mixer
				.getLine(new Line.Info(SourceDataLine.class));

		Assert.assertEquals(0, mixer.getSourceLines().length);
		sourceDataLine.open();
		Assert.assertEquals(1, mixer.getSourceLines().length);
		Assert.assertEquals(sourceDataLine, mixer.getSourceLines()[0]);
		sourceDataLine.close();
		Assert.assertEquals(0, mixer.getSourceLines().length);

	}

	@Test
	public void testAllSourceLinesClosed() {
		Assert.assertEquals(0, mixer.getSourceLines().length);

	}

	@Test
	public void testSettingStreamName() throws LineUnavailableException,
			UnsupportedAudioFileException, IOException {
		File soundFile = new File("testsounds/logout.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		AudioFormat audioFormat = audioInputStream.getFormat();

		PulseAudioSourceDataLine line;
		line = (PulseAudioSourceDataLine) mixer.getLine(new DataLine.Info(
				SourceDataLine.class, audioFormat));

		String name = "Knights Who Say ... Oh my god, i am so sorry, i didnt mean it...";
		line.setName(name);

		line.open(audioFormat);
		line.start();

		byte[] abData = new byte[1000];
		int bytesRead = 0;

		while (bytesRead >= 0) {
			bytesRead = audioInputStream.read(abData, 0, abData.length);
			if (bytesRead > 0) {
				line.write(abData, 0, bytesRead);
			}
		}

		Assert.assertTrue(line.getName() == name);
		/*
		 * FIXME test that PulseAudio also knows this correctly using
		 * introspection
		 */

		line.drain();
		line.stop();
		line.close();

	}

	@Test
	public void messWithStreams() throws LineUnavailableException {
		System.out
				.println("This test tries to unCork a stream which hasnt been corked");

		PulseAudioSourceDataLine line = (PulseAudioSourceDataLine) mixer
				.getLine(new DataLine.Info(SourceDataLine.class,
						aSupportedFormat, 1000));

		line.open();
		line.start();
		Stream s = line.getStream();
		Operation o;
		synchronized (EventLoop.getEventLoop().threadLock) {
			o = s.unCork();
		}
		o.waitForCompletion();
		o.releaseReference();
		line.stop();
		line.close();
	}

	@After
	public void tearDown() throws Exception {
		started = 0;
		stopped = 0;

		if (mixer.isOpen()) {
			mixer.close();
		}
	}

}
